@@ -36,6 +36,7 @@
 #include "shrpx_worker.h"
 #include "shrpx_http2_session.h"
 #include "shrpx_ssl.h"
+#include "shrpx_log.h"
 #include "http2.h"
 #include "util.h"
 
@@ -48,6 +49,10 @@ void timeoutcb(struct ev_loop *loop, ev_timer *w, int revents) {
   auto conn = static_cast<Connection *>(w->data);
   auto dconn = static_cast<HttpDownstreamConnection *>(conn->data);
 
+  if (w == &conn->rt && !conn->expired_rt()) {
+    return;
+  }
+
   if (LOG_ENABLED(INFO)) {
     DCLOG(INFO, dconn) << "Time out";
   }
@@ -66,6 +71,46 @@ void timeoutcb(struct ev_loop *loop, ev_timer *w, int revents) {
 }
 } // namespace
 
+namespace {
+void connect_timeoutcb(struct ev_loop *loop, ev_timer *w, int revents) {
+  auto conn = static_cast<Connection *>(w->data);
+  auto dconn = static_cast<HttpDownstreamConnection *>(conn->data);
+  auto addr = dconn->get_addr();
+  auto raddr = dconn->get_raddr();
+
+  DCLOG(WARN, dconn) << "Connect time out; addr="
+                     << util::to_numeric_addr(raddr);
+
+  downstream_failure(addr, raddr);
+
+  auto downstream = dconn->get_downstream();
+  auto upstream = downstream->get_upstream();
+  auto handler = upstream->get_client_handler();
+
+  downstream->pop_downstream_connection();
+
+  int rv;
+  auto ndconn = handler->get_downstream_connection(rv, downstream);
+  if (ndconn) {
+    if (downstream->attach_downstream_connection(std::move(ndconn)) == 0) {
+      return;
+    }
+  }
+
+  downstream->set_request_state(Downstream::CONNECT_FAIL);
+
+  if (rv == SHRPX_ERR_TLS_REQUIRED) {
+    rv = upstream->on_downstream_abort_request_with_https_redirect(downstream);
+  } else {
+    rv = upstream->on_downstream_abort_request(downstream, 504);
+  }
+
+  if (rv != 0) {
+    delete handler;
+  }
+}
+} // namespace
+
 namespace {
 void readcb(struct ev_loop *loop, ev_io *w, int revents) {
   auto conn = static_cast<Connection *>(w->data);
@@ -80,15 +125,58 @@ void readcb(struct ev_loop *loop, ev_io *w, int revents) {
 }
 } // namespace
 
+namespace {
+void backend_retry(Downstream *downstream) {
+  auto upstream = downstream->get_upstream();
+  auto handler = upstream->get_client_handler();
+
+  downstream->add_retry();
+
+  if (downstream->no_more_retry()) {
+    delete handler;
+    return;
+  }
+
+  downstream->pop_downstream_connection();
+
+  int rv;
+  auto ndconn = handler->get_downstream_connection(rv, downstream);
+  if (ndconn) {
+    if (downstream->attach_downstream_connection(std::move(ndconn)) == 0) {
+      return;
+    }
+  }
+
+  downstream->set_request_state(Downstream::CONNECT_FAIL);
+
+  if (rv == SHRPX_ERR_TLS_REQUIRED) {
+    rv = upstream->on_downstream_abort_request_with_https_redirect(downstream);
+  } else {
+    rv = upstream->on_downstream_abort_request(downstream, 503);
+  }
+
+  if (rv != 0) {
+    delete handler;
+  }
+}
+} // namespace
+
 namespace {
 void writecb(struct ev_loop *loop, ev_io *w, int revents) {
+  int rv;
   auto conn = static_cast<Connection *>(w->data);
   auto dconn = static_cast<HttpDownstreamConnection *>(conn->data);
   auto downstream = dconn->get_downstream();
   auto upstream = downstream->get_upstream();
   auto handler = upstream->get_client_handler();
 
-  if (upstream->downstream_write(dconn) != 0) {
+  rv = upstream->downstream_write(dconn);
+  if (rv == SHRPX_ERR_RETRY) {
+    backend_retry(downstream);
+    return;
+  }
+
+  if (rv != 0) {
     delete handler;
   }
 }
@@ -99,54 +187,69 @@ void connectcb(struct ev_loop *loop, ev_io *w, int revents) {
   auto conn = static_cast<Connection *>(w->data);
   auto dconn = static_cast<HttpDownstreamConnection *>(conn->data);
   auto downstream = dconn->get_downstream();
-  auto upstream = downstream->get_upstream();
-  auto handler = upstream->get_client_handler();
   if (dconn->connected() != 0) {
-    downstream->pop_downstream_connection();
-
-    auto ndconn = handler->get_downstream_connection(downstream);
-    if (ndconn) {
-      if (downstream->attach_downstream_connection(std::move(ndconn)) == 0) {
-        return;
-      }
-    }
-
-    downstream->set_request_state(Downstream::CONNECT_FAIL);
-
-    if (upstream->on_downstream_abort_request(downstream, 503) != 0) {
-      delete handler;
-    }
+    backend_retry(downstream);
     return;
   }
   writecb(loop, w, revents);
 }
 } // namespace
 
-HttpDownstreamConnection::HttpDownstreamConnection(DownstreamAddrGroup *group,
-                                                   struct ev_loop *loop,
-                                                   Worker *worker)
+HttpDownstreamConnection::HttpDownstreamConnection(
+    const std::shared_ptr<DownstreamAddrGroup> &group, ssize_t initial_addr_idx,
+    struct ev_loop *loop, Worker *worker)
     : conn_(loop, -1, nullptr, worker->get_mcpool(),
-            get_config()->conn.downstream.timeout.write,
-            get_config()->conn.downstream.timeout.read, {}, {}, connectcb,
-            readcb, timeoutcb, this, get_config()->tls.dyn_rec.warmup_threshold,
+            worker->get_downstream_config()->timeout.write,
+            worker->get_downstream_config()->timeout.read, {}, {}, connectcb,
+            readcb, connect_timeoutcb, this,
+            get_config()->tls.dyn_rec.warmup_threshold,
             get_config()->tls.dyn_rec.idle_timeout, PROTO_HTTP1),
-      do_read_(&HttpDownstreamConnection::noop),
-      do_write_(&HttpDownstreamConnection::noop),
-      do_signal_write_(&HttpDownstreamConnection::noop),
+      on_read_(&HttpDownstreamConnection::noop),
+      on_write_(&HttpDownstreamConnection::noop),
+      signal_write_(&HttpDownstreamConnection::noop),
       worker_(worker),
-      ssl_ctx_(group->shared_addr->tls ? worker->get_cl_ssl_ctx() : nullptr),
+      ssl_ctx_(worker->get_cl_ssl_ctx()),
       group_(group),
       addr_(nullptr),
+      raddr_(nullptr),
       ioctrl_(&conn_.rlimit),
-      response_htp_{0} {}
+      response_htp_{0},
+      initial_addr_idx_(initial_addr_idx),
+      reuse_first_write_done_(true),
+      reusable_(true) {}
 
-HttpDownstreamConnection::~HttpDownstreamConnection() {}
+HttpDownstreamConnection::~HttpDownstreamConnection() {
+  if (LOG_ENABLED(INFO)) {
+    DCLOG(INFO, this) << "Deleted";
+  }
+
+  if (dns_query_) {
+    auto dns_tracker = worker_->get_dns_tracker();
+    dns_tracker->cancel(dns_query_.get());
+  }
+}
 
 int HttpDownstreamConnection::attach_downstream(Downstream *downstream) {
+  int rv;
+
   if (LOG_ENABLED(INFO)) {
     DCLOG(INFO, this) << "Attaching to DOWNSTREAM:" << downstream;
   }
 
+  downstream_ = downstream;
+
+  rv = initiate_connection();
+  if (rv != 0) {
+    downstream_ = nullptr;
+    return rv;
+  }
+
+  return 0;
+}
+
+int HttpDownstreamConnection::initiate_connection() {
+  int rv;
+
   auto worker_blocker = worker_->get_connect_blocker();
   if (worker_blocker->blocked()) {
     if (LOG_ENABLED(INFO)) {
@@ -156,53 +259,131 @@ int HttpDownstreamConnection::attach_downstream(Downstream *downstream) {
     return SHRPX_ERR_NETWORK;
   }
 
-  auto &downstreamconf = get_config()->conn.downstream;
+  auto &downstreamconf = *worker_->get_downstream_config();
 
   if (conn_.fd == -1) {
-    if (ssl_ctx_) {
-      auto ssl = ssl::create_ssl(ssl_ctx_);
-      if (!ssl) {
-        return -1;
-      }
-
-      ssl::setup_downstream_http1_alpn(ssl);
-
-      conn_.set_ssl(ssl);
-    }
-
     auto &shared_addr = group_->shared_addr;
     auto &addrs = shared_addr->addrs;
-    auto &next_downstream = shared_addr->next;
+
+    // If session affinity is enabled, we always start with address at
+    // initial_addr_idx_.
+    size_t temp_idx = initial_addr_idx_;
+
+    auto &next_downstream =
+        shared_addr->affinity == AFFINITY_NONE ? shared_addr->next : temp_idx;
     auto end = next_downstream;
     for (;;) {
-      auto &addr = addrs[next_downstream];
+      auto check_dns_result = dns_query_.get() != nullptr;
 
-      if (++next_downstream >= addrs.size()) {
-        next_downstream = 0;
+      DownstreamAddr *addr;
+      if (check_dns_result) {
+        addr = addr_;
+        addr_ = nullptr;
+        assert(addr);
+        assert(addr->dns);
+      } else {
+        assert(addr_ == nullptr);
+        addr = &addrs[next_downstream];
+
+        if (++next_downstream >= addrs.size()) {
+          next_downstream = 0;
+        }
+
+        if (addr->proto != PROTO_HTTP1) {
+          if (end == next_downstream) {
+            return SHRPX_ERR_NETWORK;
+          }
+
+          continue;
+        }
       }
 
-      auto &connect_blocker = addr.connect_blocker;
+      auto &connect_blocker = addr->connect_blocker;
 
       if (connect_blocker->blocked()) {
         if (LOG_ENABLED(INFO)) {
-          DCLOG(INFO, this) << "Backend server "
-                            << util::to_numeric_addr(&addr.addr)
-                            << " was not available temporarily";
+          DCLOG(INFO, this) << "Backend server " << addr->host << ":"
+                            << addr->port << " was not available temporarily";
         }
 
-        if (end == next_downstream) {
+        if (check_dns_result) {
+          dns_query_.reset();
+        } else if (end == next_downstream) {
           return SHRPX_ERR_NETWORK;
         }
 
         continue;
       }
 
-      conn_.fd = util::create_nonblock_socket(addr.addr.su.storage.ss_family);
+      Address *raddr;
+
+      if (addr->dns) {
+        if (!check_dns_result) {
+          auto dns_query = make_unique<DNSQuery>(
+              addr->host, [this](int status, const Address *result) {
+                int rv;
+
+                if (status == DNS_STATUS_OK) {
+                  *this->resolved_addr_ = *result;
+                }
+
+                rv = this->initiate_connection();
+                if (rv != 0) {
+                  // This callback destroys |this|.
+                  auto downstream = this->downstream_;
+                  backend_retry(downstream);
+                }
+              });
+
+          auto dns_tracker = worker_->get_dns_tracker();
+
+          if (!resolved_addr_) {
+            resolved_addr_ = make_unique<Address>();
+          }
+          rv = dns_tracker->resolve(resolved_addr_.get(), dns_query.get());
+          switch (rv) {
+          case DNS_STATUS_ERROR:
+            downstream_failure(addr, nullptr);
+            if (end == next_downstream) {
+              return SHRPX_ERR_NETWORK;
+            }
+            continue;
+          case DNS_STATUS_RUNNING:
+            dns_query_ = std::move(dns_query);
+            // Remember current addr
+            addr_ = addr;
+            return 0;
+          case DNS_STATUS_OK:
+            break;
+          default:
+            assert(0);
+          }
+        } else {
+          switch (dns_query_->status) {
+          case DNS_STATUS_ERROR:
+            dns_query_.reset();
+            downstream_failure(addr, nullptr);
+            continue;
+          case DNS_STATUS_OK:
+            dns_query_.reset();
+            break;
+          default:
+            assert(0);
+          }
+        }
+
+        raddr = resolved_addr_.get();
+        util::set_port(*resolved_addr_, addr->port);
+      } else {
+        raddr = &addr->addr;
+      }
+
+      conn_.fd = util::create_nonblock_socket(raddr->su.storage.ss_family);
 
       if (conn_.fd == -1) {
         auto error = errno;
         DCLOG(WARN, this) << "socket() failed; addr="
-                          << util::to_numeric_addr(&addr.addr)
+                          << util::to_numeric_addr(raddr)
                           << ", errno=" << error;
 
         worker_blocker->on_failure();
@@ -212,19 +393,19 @@ int HttpDownstreamConnection::attach_downstream(Downstream *downstream) {
 
       worker_blocker->on_success();
 
-      int rv;
-      rv = connect(conn_.fd, &addr.addr.su.sa, addr.addr.len);
+      rv = connect(conn_.fd, &raddr->su.sa, raddr->len);
       if (rv != 0 && errno != EINPROGRESS) {
         auto error = errno;
         DCLOG(WARN, this) << "connect() failed; addr="
-                          << util::to_numeric_addr(&addr.addr)
+                          << util::to_numeric_addr(raddr)
                           << ", errno=" << error;
 
-        connect_blocker->on_failure();
+        downstream_failure(addr, raddr);
+
         close(conn_.fd);
         conn_.fd = -1;
 
-        if (end == next_downstream) {
+        if (!check_dns_result && end == next_downstream) {
           return SHRPX_ERR_NETWORK;
         }
 
@@ -236,17 +417,28 @@ int HttpDownstreamConnection::attach_downstream(Downstream *downstream) {
         DCLOG(INFO, this) << "Connecting to downstream server";
       }
 
-      addr_ = &addr;
+      addr_ = addr;
+      raddr_ = raddr;
 
-      if (ssl_ctx_) {
-        auto sni_name = !get_config()->tls.backend_sni_name.empty()
-                            ? StringRef(get_config()->tls.backend_sni_name)
-                            : StringRef(addr_->host);
+      if (addr_->tls) {
+        assert(ssl_ctx_);
+
+        auto ssl = ssl::create_ssl(ssl_ctx_);
+        if (!ssl) {
+          return -1;
+        }
+
+        ssl::setup_downstream_http1_alpn(ssl);
+
+        conn_.set_ssl(ssl);
+
+        auto sni_name =
+            addr_->sni.empty() ? StringRef{addr_->host} : StringRef{addr_->sni};
         if (!util::numeric_host(sni_name.c_str())) {
           SSL_set_tlsext_host_name(conn_.tls.ssl, sni_name.c_str());
         }
 
-        auto session = ssl::reuse_tls_session(addr_);
+        auto session = ssl::reuse_tls_session(addr_->tls_session_cache);
         if (session) {
           SSL_set_session(conn_.tls.ssl, session);
           SSL_SESSION_free(session);
@@ -263,17 +455,23 @@ int HttpDownstreamConnection::attach_downstream(Downstream *downstream) {
       break;
     }
 
-    // TODO we should have timeout for connection establishment
+    conn_.wt.repeat = downstreamconf.timeout.connect;
     ev_timer_again(conn_.loop, &conn_.wt);
   } else {
     // we may set read timer cb to idle_timeoutcb.  Reset again.
-    conn_.rt.repeat = downstreamconf.timeout.read;
     ev_set_cb(&conn_.rt, timeoutcb);
-    ev_timer_again(conn_.loop, &conn_.rt);
-    ev_set_cb(&conn_.rev, readcb);
-  }
+    if (conn_.read_timeout < downstreamconf.timeout.read) {
+      conn_.read_timeout = downstreamconf.timeout.read;
+      conn_.last_read = ev_now(conn_.loop);
+    } else {
+      conn_.again_rt(downstreamconf.timeout.read);
+    }
 
-  downstream_ = downstream;
+    ev_set_cb(&conn_.rev, readcb);
+
+    on_write_ = &HttpDownstreamConnection::write_reuse_first;
+    reuse_first_write_done_ = false;
+  }
 
   http_parser_init(&response_htp_, HTTP_RESPONSE);
   response_htp_.data = downstream_;
@@ -282,6 +480,10 @@ int HttpDownstreamConnection::attach_downstream(Downstream *downstream) {
 }
 
 int HttpDownstreamConnection::push_request_headers() {
+  if (downstream_->get_request_header_sent()) {
+    return 0;
+  }
+
   const auto &downstream_hostport = addr_->hostport;
   const auto &req = downstream_->request();
 
@@ -289,13 +491,18 @@ int HttpDownstreamConnection::push_request_headers() {
 
   auto connect_method = req.method == HTTP_CONNECT;
 
-  auto &httpconf = get_config()->http;
+  auto config = get_config();
+  auto &httpconf = config->http;
+
+  // Set request_sent to true because we write request into buffer
+  // here.
+  downstream_->set_request_header_sent(true);
 
   // For HTTP/1.0 request, there is no authority in request.  In that
   // case, we use backend server's host nonetheless.
   auto authority = StringRef(downstream_hostport);
   auto no_host_rewrite =
-      httpconf.no_host_rewrite || get_config()->http2_proxy || connect_method;
+      httpconf.no_host_rewrite || config->http2_proxy || connect_method;
 
   if (no_host_rewrite && !req.authority.empty()) {
     authority = req.authority;
@@ -308,11 +515,11 @@ int HttpDownstreamConnection::push_request_headers() {
   // Assume that method and request path do not contain \r\n.
   auto meth = http2::to_method_string(req.method);
   buf->append(meth);
-  buf->append(" ");
+  buf->append(' ');
 
   if (connect_method) {
     buf->append(authority);
-  } else if (get_config()->http2_proxy) {
+  } else if (config->http2_proxy) {
     // Construct absolute-form request target because we are going to
     // send a request to a HTTP/1 proxy.
     assert(!req.scheme.empty());
@@ -339,9 +546,9 @@ int HttpDownstreamConnection::push_request_headers() {
     buf->append("\r\n");
   }
 
-  if (!connect_method && req.http2_expect_body &&
-      !req.fs.header(http2::HD_CONTENT_LENGTH)) {
-
+  // set transfer-encoding only when content-length is unknown and
+  // request body is expected.
+  if (!connect_method && req.http2_expect_body && req.fs.content_length == -1) {
     downstream_->set_chunked_request(true);
     buf->append("Transfer-Encoding: chunked\r\n");
   }
@@ -377,7 +584,7 @@ int HttpDownstreamConnection::push_request_headers() {
   if (fwdconf.params) {
     auto params = fwdconf.params;
 
-    if (get_config()->http2_proxy || connect_method) {
+    if (config->http2_proxy || connect_method) {
       params &= ~FORWARDED_PROTO;
     }
 
@@ -421,7 +628,7 @@ int HttpDownstreamConnection::push_request_headers() {
     buf->append((*xff).value);
     buf->append("\r\n");
   }
-  if (!get_config()->http2_proxy && !connect_method) {
+  if (!config->http2_proxy && !connect_method) {
     buf->append("X-Forwarded-Proto: ");
     assert(!req.scheme.empty());
     buf->append(req.scheme);
@@ -465,10 +672,17 @@ int HttpDownstreamConnection::push_request_headers() {
       nhdrs = http::colorizeHeaders(nhdrs.c_str());
     }
     DCLOG(INFO, this) << "HTTP request headers. stream_id="
-                      << downstream_->get_stream_id() << "\n" << nhdrs;
+                      << downstream_->get_stream_id() << "\n"
+                      << nhdrs;
   }
 
-  signal_write();
+  // Don't call signal_write() if we anticipate request body.  We call
+  // signal_write() when we received request body chunk, and it
+  // enables us to send headers and data in one writev system call.
+  if (connect_method ||
+      (!req.http2_expect_body && req.fs.content_length == 0)) {
+    signal_write();
+  }
 
   return 0;
 }
@@ -496,6 +710,8 @@ int HttpDownstreamConnection::push_upload_data_chunk(const uint8_t *data,
 }
 
 int HttpDownstreamConnection::end_upload_data() {
+  signal_write();
+
   if (!downstream_->get_chunked_request()) {
     return 0;
   }
@@ -512,11 +728,27 @@ int HttpDownstreamConnection::end_upload_data() {
     output->append("\r\n");
   }
 
-  signal_write();
-
   return 0;
 }
 
+namespace {
+void remove_from_pool(HttpDownstreamConnection *dconn) {
+  auto &group = dconn->get_downstream_addr_group();
+  auto &shared_addr = group->shared_addr;
+
+  if (shared_addr->affinity == AFFINITY_NONE) {
+    auto &dconn_pool =
+        dconn->get_downstream_addr_group()->shared_addr->dconn_pool;
+    dconn_pool.remove_downstream_connection(dconn);
+    return;
+  }
+
+  auto addr = dconn->get_addr();
+  auto &dconn_pool = addr->dconn_pool;
+  dconn_pool->remove_downstream_connection(dconn);
+}
+} // namespace
+
 namespace {
 void idle_readcb(struct ev_loop *loop, ev_io *w, int revents) {
   auto conn = static_cast<Connection *>(w->data);
@@ -524,9 +756,8 @@ void idle_readcb(struct ev_loop *loop, ev_io *w, int revents) {
   if (LOG_ENABLED(INFO)) {
     DCLOG(INFO, dconn) << "Idle connection EOF";
   }
-  auto &dconn_pool =
-      dconn->get_downstream_addr_group()->shared_addr->dconn_pool;
-  dconn_pool.remove_downstream_connection(dconn);
+
+  remove_from_pool(dconn);
   // dconn was deleted
 }
 } // namespace
@@ -535,12 +766,14 @@ namespace {
 void idle_timeoutcb(struct ev_loop *loop, ev_timer *w, int revents) {
   auto conn = static_cast<Connection *>(w->data);
   auto dconn = static_cast<HttpDownstreamConnection *>(conn->data);
+
+  // We don't have to check conn->expired_rt() since we restart timer
+  // when connection gets idle.
   if (LOG_ENABLED(INFO)) {
     DCLOG(INFO, dconn) << "Idle connection timeout";
   }
-  auto &dconn_pool =
-      dconn->get_downstream_addr_group()->shared_addr->dconn_pool;
-  dconn_pool.remove_downstream_connection(dconn);
+
+  remove_from_pool(dconn);
   // dconn was deleted
 }
 } // namespace
@@ -554,9 +787,15 @@ void HttpDownstreamConnection::detach_downstream(Downstream *downstream) {
   ev_set_cb(&conn_.rev, idle_readcb);
   ioctrl_.force_resume_read();
 
-  conn_.rt.repeat = get_config()->conn.downstream.timeout.idle_read;
+  auto &downstreamconf = *worker_->get_downstream_config();
+
   ev_set_cb(&conn_.rt, idle_timeoutcb);
-  ev_timer_again(conn_.loop, &conn_.rt);
+  if (conn_.read_timeout < downstreamconf.timeout.idle_read) {
+    conn_.read_timeout = downstreamconf.timeout.idle_read;
+    conn_.last_read = ev_now(conn_.loop);
+  } else {
+    conn_.again_rt(downstreamconf.timeout.idle_read);
+  }
 
   conn_.wlimit.stopw();
   ev_timer_stop(conn_.loop, &conn_.wt);
@@ -568,8 +807,10 @@ void HttpDownstreamConnection::pause_read(IOCtrlReason reason) {
 
 int HttpDownstreamConnection::resume_read(IOCtrlReason reason,
                                           size_t consumed) {
+  auto &downstreamconf = *worker_->get_downstream_config();
+
   if (downstream_->get_response_buf()->rleft() <=
-      get_config()->conn.downstream.request_buffer_size / 2) {
+      downstreamconf.request_buffer_size / 2) {
     ioctrl_.resume_read(reason);
   }
 
@@ -596,15 +837,60 @@ namespace {
 int htp_hdrs_completecb(http_parser *htp) {
   auto downstream = static_cast<Downstream *>(htp->data);
   auto upstream = downstream->get_upstream();
+  auto handler = upstream->get_client_handler();
   const auto &req = downstream->request();
   auto &resp = downstream->response();
   int rv;
 
+  auto config = get_config();
+  auto &loggingconf = config->logging;
+
   resp.http_status = htp->status_code;
   resp.http_major = htp->http_major;
   resp.http_minor = htp->http_minor;
 
-  if (resp.fs.parse_content_length() != 0) {
+  if (resp.http_major > 1 || req.http_minor > 1) {
+    resp.http_major = 1;
+    resp.http_minor = 1;
+    return -1;
+  }
+
+  auto dconn = downstream->get_downstream_connection();
+
+  downstream->set_downstream_addr_group(dconn->get_downstream_addr_group());
+  downstream->set_addr(dconn->get_addr());
+
+  // Server MUST NOT send Transfer-Encoding with a status code 1xx or
+  // 204.  Also server MUST NOT send Transfer-Encoding with a status
+  // code 200 to a CONNECT request.  Same holds true with
+  // Content-Length.
+  if (resp.http_status == 204) {
+    if (resp.fs.header(http2::HD_TRANSFER_ENCODING)) {
+      return -1;
+    }
+    // Some server send content-length: 0 for 204.  Until they get
+    // fixed, we accept, but ignore it.
+
+    // Calling parse_content_length() detects duplicated
+    // content-length header fields.
+    if (resp.fs.parse_content_length() != 0) {
+      return -1;
+    }
+    if (resp.fs.content_length != 0) {
+      return -1;
+    }
+    if (resp.fs.content_length == 0) {
+      auto cl = resp.fs.header(http2::HD_CONTENT_LENGTH);
+      assert(cl);
+      http2::erase_header(cl);
+    }
+  } else if (resp.http_status / 100 == 1 ||
+             (resp.http_status == 200 && req.method == HTTP_CONNECT)) {
+    if (resp.fs.header(http2::HD_CONTENT_LENGTH) ||
+        resp.fs.header(http2::HD_TRANSFER_ENCODING)) {
+      return -1;
+    }
+  } else if (resp.fs.parse_content_length() != 0) {
     downstream->set_response_state(Downstream::MSG_BAD_HEADER);
     return -1;
   }
@@ -639,7 +925,15 @@ int htp_hdrs_completecb(http_parser *htp) {
     resp.connection_close = true;
     // transfer-encoding not applied to upgraded connection
     downstream->set_chunked_response(false);
+  } else if (!downstream->expect_response_body()) {
+    downstream->set_chunked_response(false);
   }
+
+  if (loggingconf.access.write_early && downstream->accesslog_ready()) {
+    handler->write_accesslog(downstream);
+    downstream->set_accesslog_written(true);
+  }
+
   if (upstream->on_downstream_header_complete(downstream) != 0) {
     return -1;
   }
@@ -656,7 +950,6 @@ int htp_hdrs_completecb(http_parser *htp) {
     }
   }
 
-  auto status = resp.http_status;
   // Ignore the response body. HEAD response may contain
   // Content-Length or Transfer-Encoding: chunked.  Some server send
   // 304 status code with nonzero Content-Length, but without response
@@ -665,10 +958,7 @@ int htp_hdrs_completecb(http_parser *htp) {
 
   // TODO It seems that the cases other than HEAD are handled by
   // http-parser.  Need test.
-  return req.method == HTTP_HEAD || (100 <= status && status <= 199) ||
-                 status == 204 || status == 304
-             ? 1
-             : 0;
+  return !http2::expect_response_body(req.method, resp.http_status);
 }
 } // namespace
 
@@ -696,8 +986,8 @@ int ensure_max_header_fields(const Downstream *downstream,
 
   if (resp.fs.num_fields() >= httpconf.max_response_header_fields) {
     if (LOG_ENABLED(INFO)) {
-      DLOG(INFO, downstream)
-          << "Too many header field num=" << resp.fs.num_fields() + 1;
+      DLOG(INFO, downstream) << "Too many header field num="
+                             << resp.fs.num_fields() + 1;
     }
     return -1;
   }
@@ -711,7 +1001,6 @@ int htp_hdr_keycb(http_parser *htp, const char *data, size_t len) {
   auto downstream = static_cast<Downstream *>(htp->data);
   auto &resp = downstream->response();
   auto &httpconf = get_config()->http;
-  auto &balloc = downstream->get_block_allocator();
 
   if (ensure_header_field_buffer(downstream, httpconf, len) != 0) {
     return -1;
@@ -724,9 +1013,7 @@ int htp_hdr_keycb(http_parser *htp, const char *data, size_t len) {
       if (ensure_max_header_fields(downstream, httpconf) != 0) {
         return -1;
       }
-      auto name = http2::copy_lower(balloc, StringRef{data, len});
-      auto token = http2::lookup_token(name);
-      resp.fs.add_header_token(name, StringRef{}, false, token);
+      resp.fs.alloc_add_header_name(StringRef{data, len});
     }
   } else {
     // trailer part
@@ -739,9 +1026,7 @@ int htp_hdr_keycb(http_parser *htp, const char *data, size_t len) {
         // wrong place or crash if trailer fields are currently empty.
         return -1;
       }
-      auto name = http2::copy_lower(balloc, StringRef{data, len});
-      auto token = http2::lookup_token(name);
-      resp.fs.add_trailer_token(name, StringRef{}, false, token);
+      resp.fs.alloc_add_trailer_name(StringRef{data, len});
     }
   }
   return 0;
@@ -808,7 +1093,7 @@ int htp_msg_completecb(http_parser *htp) {
 } // namespace
 
 namespace {
-http_parser_settings htp_hooks = {
+constexpr http_parser_settings htp_hooks = {
     htp_msg_begincb,     // http_cb on_message_begin;
     nullptr,             // http_data_cb on_url;
     nullptr,             // http_data_cb on_status;
@@ -820,8 +1105,33 @@ http_parser_settings htp_hooks = {
 };
 } // namespace
 
+int HttpDownstreamConnection::write_reuse_first() {
+  int rv;
+
+  if (conn_.tls.ssl) {
+    rv = write_tls();
+  } else {
+    rv = write_clear();
+  }
+
+  if (rv != 0) {
+    return SHRPX_ERR_RETRY;
+  }
+
+  if (conn_.tls.ssl) {
+    on_write_ = &HttpDownstreamConnection::write_tls;
+  } else {
+    on_write_ = &HttpDownstreamConnection::write_clear;
+  }
+
+  reuse_first_write_done_ = true;
+
+  return 0;
+}
+
 int HttpDownstreamConnection::read_clear() {
-  ev_timer_again(conn_.loop, &conn_.rt);
+  conn_.last_read = ev_now(conn_.loop);
+
   std::array<uint8_t, 16_k> buf;
   int rv;
 
@@ -847,7 +1157,7 @@ int HttpDownstreamConnection::read_clear() {
 }
 
 int HttpDownstreamConnection::write_clear() {
-  ev_timer_again(conn_.loop, &conn_.rt);
+  conn_.last_read = ev_now(conn_.loop);
 
   auto upstream = downstream_->get_upstream();
   auto input = downstream_->get_request_buf();
@@ -864,7 +1174,16 @@ int HttpDownstreamConnection::write_clear() {
     }
 
     if (nwrite < 0) {
-      return nwrite;
+      if (!reuse_first_write_done_) {
+        return nwrite;
+      }
+      // We may have pending data in receive buffer which may contain
+      // part of response body.  So keep reading.  Invoke read event
+      // to get read(2) error just in case.
+      ev_feed_event(conn_.loop, &conn_.rev, EV_READ);
+      on_write_ = &HttpDownstreamConnection::noop;
+      reusable_ = false;
+      break;
     }
 
     input->drain(nwrite);
@@ -886,7 +1205,7 @@ int HttpDownstreamConnection::write_clear() {
 int HttpDownstreamConnection::tls_handshake() {
   ERR_clear_error();
 
-  ev_timer_again(conn_.loop, &conn_.rt);
+  conn_.last_read = ev_now(conn_.loop);
 
   auto rv = conn_.tls_handshake();
   if (rv == SHRPX_ERR_INPROGRESS) {
@@ -894,6 +1213,8 @@ int HttpDownstreamConnection::tls_handshake() {
   }
 
   if (rv < 0) {
+    downstream_failure(addr_, raddr_);
+
     return rv;
   }
 
@@ -902,19 +1223,31 @@ int HttpDownstreamConnection::tls_handshake() {
   }
 
   if (!get_config()->tls.insecure &&
-      ssl::check_cert(conn_.tls.ssl, addr_) != 0) {
+      ssl::check_cert(conn_.tls.ssl, addr_, raddr_) != 0) {
+    downstream_failure(addr_, raddr_);
+
     return -1;
   }
 
   if (!SSL_session_reused(conn_.tls.ssl)) {
     auto session = SSL_get0_session(conn_.tls.ssl);
     if (session) {
-      ssl::try_cache_tls_session(addr_, session, ev_now(conn_.loop));
+      ssl::try_cache_tls_session(addr_->tls_session_cache, *raddr_, session,
+                                 ev_now(conn_.loop));
     }
   }
 
-  do_read_ = &HttpDownstreamConnection::read_tls;
-  do_write_ = &HttpDownstreamConnection::write_tls;
+  auto &connect_blocker = addr_->connect_blocker;
+
+  signal_write_ = &HttpDownstreamConnection::actual_signal_write;
+
+  connect_blocker->on_success();
+
+  ev_set_cb(&conn_.rt, timeoutcb);
+  ev_set_cb(&conn_.wt, timeoutcb);
+
+  on_read_ = &HttpDownstreamConnection::read_tls;
+  on_write_ = &HttpDownstreamConnection::write_tls;
 
   // TODO Check negotiated ALPN
 
@@ -922,9 +1255,10 @@ int HttpDownstreamConnection::tls_handshake() {
 }
 
 int HttpDownstreamConnection::read_tls() {
+  conn_.last_read = ev_now(conn_.loop);
+
   ERR_clear_error();
 
-  ev_timer_again(conn_.loop, &conn_.rt);
   std::array<uint8_t, 16_k> buf;
   int rv;
 
@@ -950,9 +1284,9 @@ int HttpDownstreamConnection::read_tls() {
 }
 
 int HttpDownstreamConnection::write_tls() {
-  ERR_clear_error();
+  conn_.last_read = ev_now(conn_.loop);
 
-  ev_timer_again(conn_.loop, &conn_.rt);
+  ERR_clear_error();
 
   auto upstream = downstream_->get_upstream();
   auto input = downstream_->get_request_buf();
@@ -969,7 +1303,16 @@ int HttpDownstreamConnection::write_tls() {
     }
 
     if (nwrite < 0) {
-      return nwrite;
+      if (!reuse_first_write_done_) {
+        return nwrite;
+      }
+      // We may have pending data in receive buffer which may contain
+      // part of response body.  So keep reading.  Invoke read event
+      // to get read(2) error just in case.
+      ev_feed_event(conn_.loop, &conn_.rev, EV_READ);
+      on_write_ = &HttpDownstreamConnection::noop;
+      reusable_ = false;
+      break;
     }
 
     input->drain(nwrite);
@@ -1058,15 +1401,15 @@ int HttpDownstreamConnection::process_input(const uint8_t *data,
 int HttpDownstreamConnection::connected() {
   auto &connect_blocker = addr_->connect_blocker;
 
-  if (!util::check_socket_connected(conn_.fd)) {
+  auto sock_error = util::get_socket_error(conn_.fd);
+  if (sock_error != 0) {
     conn_.wlimit.stopw();
 
-    if (LOG_ENABLED(INFO)) {
-      DCLOG(INFO, this) << "Backend connect failed; addr="
-                        << util::to_numeric_addr(&addr_->addr);
-    }
+    DCLOG(WARN, this) << "Backend connect failed; addr="
+                      << util::to_numeric_addr(raddr_)
+                      << ": errno=" << sock_error;
 
-    downstream_failure(addr_);
+    downstream_failure(addr_, raddr_);
 
     return -1;
   }
@@ -1075,34 +1418,44 @@ int HttpDownstreamConnection::connected() {
     DCLOG(INFO, this) << "Connected to downstream host";
   }
 
-  connect_blocker->on_success();
+  auto &downstreamconf = *get_config()->conn.downstream;
+
+  // Reset timeout for write.  Previously, we set timeout for connect.
+  conn_.wt.repeat = downstreamconf.timeout.write;
+  ev_timer_again(conn_.loop, &conn_.wt);
 
   conn_.rlimit.startw();
+  conn_.again_rt();
 
   ev_set_cb(&conn_.wev, writecb);
 
-  do_signal_write_ = &HttpDownstreamConnection::actual_signal_write;
-
   if (conn_.tls.ssl) {
-    do_read_ = &HttpDownstreamConnection::tls_handshake;
-    do_write_ = &HttpDownstreamConnection::tls_handshake;
+    on_read_ = &HttpDownstreamConnection::tls_handshake;
+    on_write_ = &HttpDownstreamConnection::tls_handshake;
 
     return 0;
   }
 
-  do_read_ = &HttpDownstreamConnection::read_clear;
-  do_write_ = &HttpDownstreamConnection::write_clear;
+  signal_write_ = &HttpDownstreamConnection::actual_signal_write;
+
+  connect_blocker->on_success();
+
+  ev_set_cb(&conn_.rt, timeoutcb);
+  ev_set_cb(&conn_.wt, timeoutcb);
+
+  on_read_ = &HttpDownstreamConnection::read_clear;
+  on_write_ = &HttpDownstreamConnection::write_clear;
 
   return 0;
 }
 
-int HttpDownstreamConnection::on_read() { return do_read_(*this); }
+int HttpDownstreamConnection::on_read() { return on_read_(*this); }
 
-int HttpDownstreamConnection::on_write() { return do_write_(*this); }
+int HttpDownstreamConnection::on_write() { return on_write_(*this); }
 
 void HttpDownstreamConnection::on_upstream_change(Upstream *upstream) {}
 
-void HttpDownstreamConnection::signal_write() { do_signal_write_(*this); }
+void HttpDownstreamConnection::signal_write() { signal_write_(*this); }
 
 int HttpDownstreamConnection::actual_signal_write() {
   ev_feed_event(conn_.loop, &conn_.wev, EV_WRITE);
@@ -1111,9 +1464,17 @@ int HttpDownstreamConnection::actual_signal_write() {
 
 int HttpDownstreamConnection::noop() { return 0; }
 
-DownstreamAddrGroup *
+const std::shared_ptr<DownstreamAddrGroup> &
 HttpDownstreamConnection::get_downstream_addr_group() const {
   return group_;
 }
 
+DownstreamAddr *HttpDownstreamConnection::get_addr() const { return addr_; }
+
+bool HttpDownstreamConnection::poolable() const {
+  return !group_->retired && reusable_;
+}
+
+const Address *HttpDownstreamConnection::get_raddr() const { return raddr_; }
+
 } // namespace shrpx
