@@ -41,6 +41,7 @@
 #endif // HAVE_MRUBY
 #include "shrpx_worker.h"
 #include "shrpx_http2_session.h"
+#include "shrpx_log.h"
 #include "http2.h"
 #include "util.h"
 #include "template.h"
@@ -105,7 +106,7 @@ ssize_t recv_callback(spdylay_session *session, uint8_t *buf, size_t len,
 
   auto nread = std::min(rb->rleft(), len);
 
-  memcpy(buf, rb->pos, nread);
+  memcpy(buf, rb->pos(), nread);
   rb->drain(nread);
   rlimit->startw();
 
@@ -136,7 +137,7 @@ void on_stream_close_callback(spdylay_session *session, int32_t stream_id,
 
   if (downstream->get_request_state() == Downstream::CONNECT_FAIL) {
     upstream->remove_downstream(downstream);
-    // downstrea was deleted
+    // downstream was deleted
 
     return;
   }
@@ -153,7 +154,7 @@ void on_stream_close_callback(spdylay_session *session, int32_t stream_id,
   // If shrpx_downstream::push_request_headers() failed, the
   // error is handled here.
   upstream->remove_downstream(downstream);
-  // downstrea was deleted
+  // downstream was deleted
 
   // How to test this case? Request sufficient large download
   // and make client send RST_STREAM after it gets first DATA
@@ -165,6 +166,7 @@ namespace {
 void on_ctrl_recv_callback(spdylay_session *session, spdylay_frame_type type,
                            spdylay_frame *frame, void *user_data) {
   auto upstream = static_cast<SpdyUpstream *>(user_data);
+  auto config = get_config();
 
   switch (type) {
   case SPDYLAY_SYN_STREAM: {
@@ -180,6 +182,10 @@ void on_ctrl_recv_callback(spdylay_session *session, spdylay_frame_type type,
 
     auto &balloc = downstream->get_block_allocator();
 
+    auto lgconf = log_config();
+    lgconf->update_tstamp(std::chrono::system_clock::now());
+    req.tstamp = lgconf->tstamp;
+
     downstream->reset_upstream_rtimer();
 
     auto nv = frame->syn_stream.nv;
@@ -190,7 +196,8 @@ void on_ctrl_recv_callback(spdylay_session *session, spdylay_frame_type type,
         ss << TTY_HTTP_HD << nv[i] << TTY_RST << ": " << nv[i + 1] << "\n";
       }
       ULOG(INFO, upstream) << "HTTP request headers. stream_id="
-                           << downstream->get_stream_id() << "\n" << ss.str();
+                           << downstream->get_stream_id() << "\n"
+                           << ss.str();
     }
 
     size_t num_headers = 0;
@@ -202,7 +209,7 @@ void on_ctrl_recv_callback(spdylay_session *session, spdylay_frame_type type,
       header_buffer += strlen(nv[i]) + strlen(nv[i + 1]);
     }
 
-    auto &httpconf = get_config()->http;
+    auto &httpconf = config->http;
 
     // spdy does not define usage of trailer fields, and we ignores
     // them.
@@ -254,9 +261,11 @@ void on_ctrl_recv_callback(spdylay_session *session, spdylay_frame_type type,
       return;
     }
 
-    if (std::find_if(std::begin(host->value), std::end(host->value),
+    auto authority = is_connect ? path : host;
+
+    if (std::find_if(std::begin(authority->value), std::end(authority->value),
                      [](char c) { return c == '"' || c == '\\'; }) !=
-        std::end(host->value)) {
+        std::end(authority->value)) {
       if (upstream->error_reply(downstream, 400) != 0) {
         ULOG(FATAL, upstream) << "error_reply failed";
       }
@@ -293,7 +302,7 @@ void on_ctrl_recv_callback(spdylay_session *session, spdylay_frame_type type,
       auto handler = upstream->get_client_handler();
       auto faddr = handler->get_upstream_addr();
 
-      if (get_config()->http2_proxy && !faddr->alt_mode) {
+      if (config->http2_proxy && !faddr->alt_mode) {
         req.path = path->value;
       } else if (method_token == HTTP_OPTIONS &&
                  path->value == StringRef::from_lit("*")) {
@@ -363,7 +372,7 @@ void SpdyUpstream::start_downstream(Downstream *downstream) {
 void SpdyUpstream::initiate_downstream(Downstream *downstream) {
   int rv;
 
-  auto dconn = handler_->get_downstream_connection(downstream);
+  auto dconn = handler_->get_downstream_connection(rv, downstream);
 
   if (!dconn ||
       (rv = downstream->attach_downstream_connection(std::move(dconn))) != 0) {
@@ -584,7 +593,8 @@ SpdyUpstream::SpdyUpstream(uint16_t version, ClientHandler *handler)
                                   &max_buffer, sizeof(max_buffer));
   assert(rv == 0);
 
-  auto &http2conf = get_config()->http2;
+  auto config = get_config();
+  auto &http2conf = config->http2;
 
   auto faddr = handler_->get_upstream_addr();
 
@@ -632,7 +642,7 @@ SpdyUpstream::SpdyUpstream(uint16_t version, ClientHandler *handler)
   }
 
   handler_->reset_upstream_read_timeout(
-      get_config()->conn.upstream.timeout.http2_read);
+      config->conn.upstream.timeout.http2_read);
 
   handler_->signal_write();
 }
@@ -740,7 +750,7 @@ int SpdyUpstream::downstream_write(DownstreamConnection *dconn) {
     return downstream_error(dconn, Downstream::EVENT_ERROR);
   }
   if (rv != 0) {
-    return -1;
+    return rv;
   }
   return 0;
 }
@@ -846,7 +856,7 @@ int SpdyUpstream::rst_stream(Downstream *downstream, int status_code) {
   if (rv < SPDYLAY_ERR_FATAL) {
     ULOG(FATAL, this) << "spdylay_submit_rst_stream() failed: "
                       << spdylay_strerror(rv);
-    DIE();
+    return -1;
   }
   return 0;
 }
@@ -870,8 +880,8 @@ ssize_t spdy_data_read_callback(spdylay_session *session, int32_t stream_id,
     } else {
       // For tunneling, issue RST_STREAM to finish the stream.
       if (LOG_ENABLED(INFO)) {
-        ULOG(INFO, upstream)
-            << "RST_STREAM to tunneled stream stream_id=" << stream_id;
+        ULOG(INFO, upstream) << "RST_STREAM to tunneled stream stream_id="
+                             << stream_id;
       }
       upstream->rst_stream(
           downstream, infer_upstream_rst_stream_status_code(
@@ -915,11 +925,12 @@ int SpdyUpstream::send_reply(Downstream *downstream, const uint8_t *body,
   const auto &resp = downstream->response();
   auto &balloc = downstream->get_block_allocator();
 
-  auto status_string = http2::get_status_string(balloc, resp.http_status);
+  auto status_line = http2::stringify_status(balloc, resp.http_status);
 
   const auto &headers = resp.fs.headers();
 
-  auto &httpconf = get_config()->http;
+  auto config = get_config();
+  auto &httpconf = config->http;
 
   auto nva = std::vector<const char *>();
   // 6 for :status, :version and server.  1 for last terminal nullptr.
@@ -927,7 +938,7 @@ int SpdyUpstream::send_reply(Downstream *downstream, const uint8_t *body,
               httpconf.add_response_headers.size() * 2 + 1);
 
   nva.push_back(":status");
-  nva.push_back(status_string.c_str());
+  nva.push_back(status_line.c_str());
   nva.push_back(":version");
   nva.push_back("HTTP/1.1");
 
@@ -948,7 +959,7 @@ int SpdyUpstream::send_reply(Downstream *downstream, const uint8_t *body,
 
   if (!resp.fs.header(http2::HD_SERVER)) {
     nva.push_back("server");
-    nva.push_back(get_config()->http.server_name.c_str());
+    nva.push_back(config->http.server_name.c_str());
   }
 
   for (auto &p : httpconf.add_response_headers) {
@@ -972,6 +983,10 @@ int SpdyUpstream::send_reply(Downstream *downstream, const uint8_t *body,
 
   downstream->set_response_state(Downstream::MSG_COMPLETE);
 
+  if (data_prd_ptr) {
+    downstream->reset_upstream_wtimer();
+  }
+
   return 0;
 }
 
@@ -995,13 +1010,15 @@ int SpdyUpstream::error_reply(Downstream *downstream,
   lgconf->update_tstamp(std::chrono::system_clock::now());
 
   auto content_length = util::make_string_ref_uint(balloc, html.size());
-  auto status_string = http2::get_status_string(balloc, status_code);
+  auto status_line = http2::stringify_status(balloc, status_code);
 
-  const char *nv[] = {":status", status_string.c_str(), ":version", "http/1.1",
-                      "content-type", "text/html; charset=UTF-8", "server",
-                      get_config()->http.server_name.c_str(), "content-length",
-                      content_length.c_str(), "date",
-                      lgconf->time_http_str.c_str(), nullptr};
+  const char *nv[] = {":status",        status_line.c_str(),
+                      ":version",       "http/1.1",
+                      "content-type",   "text/html; charset=UTF-8",
+                      "server",         get_config()->http.server_name.c_str(),
+                      "content-length", content_length.c_str(),
+                      "date",           lgconf->tstamp->time_http.c_str(),
+                      nullptr};
 
   rv = spdylay_submit_response(session_, downstream->get_stream_id(), nv,
                                &data_prd);
@@ -1011,6 +1028,8 @@ int SpdyUpstream::error_reply(Downstream *downstream,
     return -1;
   }
 
+  downstream->reset_upstream_wtimer();
+
   return 0;
 }
 
@@ -1084,9 +1103,10 @@ int SpdyUpstream::on_downstream_header_complete(Downstream *downstream) {
     DLOG(INFO, downstream) << "HTTP response header completed";
   }
 
-  auto &httpconf = get_config()->http;
+  auto config = get_config();
+  auto &httpconf = config->http;
 
-  if (!get_config()->http2_proxy && !httpconf.no_location_rewrite) {
+  if (!config->http2_proxy && !httpconf.no_location_rewrite) {
     downstream->rewrite_location_response_header(req.scheme);
   }
 
@@ -1097,9 +1117,10 @@ int SpdyUpstream::on_downstream_header_complete(Downstream *downstream) {
 
   size_t hdidx = 0;
   std::string via_value;
-  auto status_string = http2::get_status_string(balloc, resp.http_status);
+  auto status_line = http2::stringify_status(balloc, resp.http_status);
+
   nv[hdidx++] = ":status";
-  nv[hdidx++] = status_string.c_str();
+  nv[hdidx++] = status_line.c_str();
   nv[hdidx++] = ":version";
   nv[hdidx++] = "HTTP/1.1";
   for (auto &hd : resp.fs.headers()) {
@@ -1162,7 +1183,8 @@ int SpdyUpstream::on_downstream_header_complete(Downstream *downstream) {
       ss << TTY_HTTP_HD << nv[i] << TTY_RST << ": " << nv[i + 1] << "\n";
     }
     ULOG(INFO, this) << "HTTP response headers. stream_id="
-                     << downstream->get_stream_id() << "\n" << ss.str();
+                     << downstream->get_stream_id() << "\n"
+                     << ss.str();
   }
   spdylay_data_provider data_prd;
   data_prd.source.ptr = downstream;
@@ -1176,6 +1198,8 @@ int SpdyUpstream::on_downstream_header_complete(Downstream *downstream) {
     return -1;
   }
 
+  downstream->reset_upstream_wtimer();
+
   return 0;
 }
 
@@ -1251,6 +1275,13 @@ int SpdyUpstream::on_downstream_abort_request(Downstream *downstream,
   return 0;
 }
 
+int SpdyUpstream::on_downstream_abort_request_with_https_redirect(
+    Downstream *downstream) {
+  // This should not be called since SPDY is only available with TLS.
+  assert(0);
+  return 0;
+}
+
 int SpdyUpstream::consume(int32_t stream_id, size_t len) {
   int rv;
 
@@ -1277,6 +1308,8 @@ int SpdyUpstream::on_timeout(Downstream *downstream) {
 
   rst_stream(downstream, SPDYLAY_INTERNAL_ERROR);
 
+  handler_->signal_write();
+
   return 0;
 }
 
@@ -1305,6 +1338,11 @@ int SpdyUpstream::on_downstream_reset(Downstream *downstream, bool no_retry) {
   }
 
   if (!downstream->request_submission_ready()) {
+    if (downstream->get_response_state() == Downstream::MSG_COMPLETE) {
+      // We have got all response body already.  Send it off.
+      downstream->pop_downstream_connection();
+      return 0;
+    }
     rst_stream(downstream, SPDYLAY_INTERNAL_ERROR);
     downstream->pop_downstream_connection();
 
@@ -1326,7 +1364,7 @@ int SpdyUpstream::on_downstream_reset(Downstream *downstream, bool no_retry) {
   // downstream connection is clean; we can retry with new
   // downstream connection.
 
-  dconn = handler_->get_downstream_connection(downstream);
+  dconn = handler_->get_downstream_connection(rv, downstream);
   if (!dconn) {
     goto fail;
   }
