@@ -76,7 +76,8 @@ bool match_shared_downstream_addr(
     return false;
   }
 
-  if (lhs->affinity != rhs->affinity) {
+  if (lhs->affinity != rhs->affinity ||
+      lhs->redirect_if_not_tls != rhs->redirect_if_not_tls) {
     return false;
   }
 
@@ -92,7 +93,7 @@ bool match_shared_downstream_addr(
       auto &b = rhs->addrs[i];
       if (a.host == b.host && a.port == b.port && a.host_unix == b.host_unix &&
           a.proto == b.proto && a.tls == b.tls && a.sni == b.sni &&
-          a.fall == b.fall && a.rise == b.rise) {
+          a.fall == b.fall && a.rise == b.rise && a.dns == b.dns) {
         break;
       }
     }
@@ -108,18 +109,15 @@ bool match_shared_downstream_addr(
 }
 } // namespace
 
-namespace {
-std::random_device rd;
-} // namespace
-
 Worker::Worker(struct ev_loop *loop, SSL_CTX *sv_ssl_ctx, SSL_CTX *cl_ssl_ctx,
                SSL_CTX *tls_session_cache_memcached_ssl_ctx,
                ssl::CertLookupTree *cert_tree,
                const std::shared_ptr<TicketKeys> &ticket_keys,
                ConnectionHandler *conn_handler,
                std::shared_ptr<DownstreamConfig> downstreamconf)
-    : randgen_(rd()),
+    : randgen_(util::make_mt19937()),
       worker_stat_{},
+      dns_tracker_(loop),
       loop_(loop),
       sv_ssl_ctx_(sv_ssl_ctx),
       cl_ssl_ctx_(cl_ssl_ctx),
@@ -182,13 +180,15 @@ void Worker::replace_downstream_config(
     auto &dst = downstream_addr_groups_[i];
 
     dst = std::make_shared<DownstreamAddrGroup>();
-    dst->pattern = src.pattern;
+    dst->pattern =
+        ImmutableString{std::begin(src.pattern), std::end(src.pattern)};
 
     auto shared_addr = std::make_shared<SharedDownstreamAddr>();
 
     shared_addr->addrs.resize(src.addrs.size());
     shared_addr->affinity = src.affinity;
     shared_addr->affinity_hash = src.affinity_hash;
+    shared_addr->redirect_if_not_tls = src.redirect_if_not_tls;
 
     size_t num_http1 = 0;
     size_t num_http2 = 0;
@@ -198,15 +198,17 @@ void Worker::replace_downstream_config(
       auto &dst_addr = shared_addr->addrs[j];
 
       dst_addr.addr = src_addr.addr;
-      dst_addr.host = src_addr.host;
-      dst_addr.hostport = src_addr.hostport;
+      dst_addr.host = make_string_ref(shared_addr->balloc, src_addr.host);
+      dst_addr.hostport =
+          make_string_ref(shared_addr->balloc, src_addr.hostport);
       dst_addr.port = src_addr.port;
       dst_addr.host_unix = src_addr.host_unix;
       dst_addr.proto = src_addr.proto;
       dst_addr.tls = src_addr.tls;
-      dst_addr.sni = src_addr.sni;
+      dst_addr.sni = make_string_ref(shared_addr->balloc, src_addr.sni);
       dst_addr.fall = src_addr.fall;
       dst_addr.rise = src_addr.rise;
+      dst_addr.dns = src_addr.dns;
 
       auto shared_addr_ptr = shared_addr.get();
 
@@ -305,9 +307,9 @@ void Worker::wait() {
 void Worker::run_async() {
 #ifndef NOTHREADS
   fut_ = std::async(std::launch::async, [this] {
-    (void)reopen_log_files();
+    (void)reopen_log_files(get_config()->logging);
     ev_run(loop_);
-    delete log_config();
+    delete_log_config();
   });
 #endif // !NOTHREADS
 }
@@ -343,7 +345,9 @@ void Worker::process_events() {
 
   ev_timer_start(loop_, &proc_wev_timer_);
 
-  auto worker_connections = get_config()->conn.upstream.worker_connections;
+  auto config = get_config();
+
+  auto worker_connections = config->conn.upstream.worker_connections;
 
   switch (wev.type) {
   case NEW_CONNECTION: {
@@ -384,7 +388,7 @@ void Worker::process_events() {
     WLOG(NOTICE, this) << "Reopening log files: worker process (thread " << this
                        << ")";
 
-    reopen_log_files();
+    reopen_log_files(config->logging);
 
     break;
   case GRACEFUL_SHUTDOWN:
@@ -488,6 +492,8 @@ ConnectionHandler *Worker::get_connection_handler() const {
   return conn_handler_;
 }
 
+DNSTracker *Worker::get_dns_tracker() { return &dns_tracker_; }
+
 namespace {
 size_t match_downstream_addr_group_host(
     const RouterConfig &routerconf, const StringRef &host,
@@ -633,7 +639,7 @@ size_t match_downstream_addr_group(
                                           catch_all, balloc);
 }
 
-void downstream_failure(DownstreamAddr *addr) {
+void downstream_failure(DownstreamAddr *addr, const Address *raddr) {
   const auto &connect_blocker = addr->connect_blocker;
 
   if (connect_blocker->in_offline()) {
@@ -649,8 +655,15 @@ void downstream_failure(DownstreamAddr *addr) {
   auto fail_count = connect_blocker->get_fail_count();
 
   if (fail_count >= addr->fall) {
-    LOG(WARN) << "Could not connect to " << util::to_numeric_addr(&addr->addr)
-              << " " << fail_count << " times in a row; considered as offline";
+    if (raddr) {
+      LOG(WARN) << "Could not connect to " << util::to_numeric_addr(raddr)
+                << " " << fail_count
+                << " times in a row; considered as offline";
+    } else {
+      LOG(WARN) << "Could not connect to " << addr->host << ":" << addr->port
+                << " " << fail_count
+                << " times in a row; considered as offline";
+    }
 
     connect_blocker->offline();
 
