@@ -38,32 +38,15 @@ namespace nghttp2 {
 
 namespace ssl {
 
-// Recommended general purpose "Intermediate compatibility" cipher
-// suites by mozilla.
-//
-// https://wiki.mozilla.org/Security/Server_Side_TLS
-const char *const DEFAULT_CIPHER_LIST =
-    "ECDHE-ECDSA-CHACHA20-POLY1305:ECDHE-RSA-CHACHA20-POLY1305:ECDHE-ECDSA-"
-    "AES128-GCM-SHA256:ECDHE-RSA-AES128-GCM-SHA256:ECDHE-ECDSA-AES256-GCM-"
-    "SHA384:ECDHE-RSA-AES256-GCM-SHA384:DHE-RSA-AES128-GCM-SHA256:DHE-RSA-"
-    "AES256-GCM-SHA384:ECDHE-ECDSA-AES128-SHA256:ECDHE-RSA-AES128-SHA256:ECDHE-"
-    "ECDSA-AES128-SHA:ECDHE-RSA-AES256-SHA384:ECDHE-RSA-AES128-SHA:ECDHE-ECDSA-"
-    "AES256-SHA384:ECDHE-ECDSA-AES256-SHA:ECDHE-RSA-AES256-SHA:DHE-RSA-AES128-"
-    "SHA256:DHE-RSA-AES128-SHA:DHE-RSA-AES256-SHA256:DHE-RSA-AES256-SHA:ECDHE-"
-    "ECDSA-DES-CBC3-SHA:ECDHE-RSA-DES-CBC3-SHA:EDH-RSA-DES-CBC3-SHA:AES128-GCM-"
-    "SHA256:AES256-GCM-SHA384:AES128-SHA256:AES256-SHA256:AES128-SHA:AES256-"
-    "SHA:DES-CBC3-SHA:!DSS";
-
-#if OPENSSL_101_API
+#if OPENSSL_1_1_API
 
 // CRYPTO_LOCK is deprecated as of OpenSSL 1.1.0
 LibsslGlobalLock::LibsslGlobalLock() {}
-LibsslGlobalLock::~LibsslGlobalLock() {}
 
-#else // !OPENSSL_101_API
+#else // !OPENSSL_1_1_API
 
 namespace {
-std::vector<std::mutex> ssl_global_locks;
+std::mutex *ssl_global_locks;
 } // namespace
 
 namespace {
@@ -77,11 +60,11 @@ void ssl_locking_cb(int mode, int type, const char *file, int line) {
 } // namespace
 
 LibsslGlobalLock::LibsslGlobalLock() {
-  if (!ssl_global_locks.empty()) {
+  if (ssl_global_locks) {
     std::cerr << "OpenSSL global lock has been already set" << std::endl;
     assert(0);
   }
-  ssl_global_locks = std::vector<std::mutex>(CRYPTO_num_locks());
+  ssl_global_locks = new std::mutex[CRYPTO_num_locks()];
   // CRYPTO_set_id_callback(ssl_thread_id); OpenSSL manual says that
   // if threadid_func is not specified using
   // CRYPTO_THREADID_set_callback(), then default implementation is
@@ -89,9 +72,7 @@ LibsslGlobalLock::LibsslGlobalLock() {
   CRYPTO_set_locking_callback(ssl_locking_cb);
 }
 
-LibsslGlobalLock::~LibsslGlobalLock() { ssl_global_locks.clear(); }
-
-#endif // !OPENSSL_101_API
+#endif // !OPENSSL_1_1_API
 
 const char *get_tls_protocol(SSL *ssl) {
   switch (SSL_version(ssl)) {
@@ -99,6 +80,10 @@ const char *get_tls_protocol(SSL *ssl) {
     return "SSLv2";
   case SSL3_VERSION:
     return "SSLv3";
+#ifdef TLS1_3_VERSION
+  case TLS1_3_VERSION:
+    return "TLSv1.3";
+#endif // TLS1_3_VERSION
   case TLS1_2_VERSION:
     return "TLSv1.2";
   case TLS1_1_VERSION:
@@ -156,7 +141,7 @@ bool check_http2_cipher_black_list(SSL *ssl) {
 bool check_http2_tls_version(SSL *ssl) {
   auto tls_ver = SSL_version(ssl);
 
-  return tls_ver == TLS1_2_VERSION;
+  return tls_ver >= TLS1_2_VERSION;
 }
 
 bool check_http2_requirement(SSL *ssl) {
@@ -164,15 +149,51 @@ bool check_http2_requirement(SSL *ssl) {
 }
 
 void libssl_init() {
-// OPENSSL_config() is not available in BoringSSL.  It is also
-// deprecated as of OpenSSL 1.1.0.
-#if !defined(OPENSSL_IS_BORINGSSL) && !OPENSSL_101_API
+#if OPENSSL_1_1_API
+// No explicit initialization is required.
+#elif defined(OPENSSL_IS_BORINGSSL)
+  CRYPTO_library_init();
+#else  // !OPENSSL_1_1_API && !defined(OPENSSL_IS_BORINGSSL)
   OPENSSL_config(nullptr);
-#endif // !defined(OPENSSL_IS_BORINGSSL) && !OPENSSL_101_API
-
   SSL_load_error_strings();
   SSL_library_init();
   OpenSSL_add_all_algorithms();
+#endif // !OPENSSL_1_1_API && !defined(OPENSSL_IS_BORINGSSL)
+}
+
+int ssl_ctx_set_proto_versions(SSL_CTX *ssl_ctx, int min, int max) {
+#if OPENSSL_1_1_API || defined(OPENSSL_IS_BORINGSSL)
+  if (SSL_CTX_set_min_proto_version(ssl_ctx, min) != 1 ||
+      SSL_CTX_set_max_proto_version(ssl_ctx, max) != 1) {
+    return -1;
+  }
+  return 0;
+#else  // !OPENSSL_1_1_API && !defined(OPENSSL_IS_BORINGSSL)
+  long int opts = 0;
+
+  // TODO We depends on the ordering of protocol version macro in
+  // OpenSSL.
+  if (min > TLS1_VERSION) {
+    opts |= SSL_OP_NO_TLSv1;
+  }
+  if (min > TLS1_1_VERSION) {
+    opts |= SSL_OP_NO_TLSv1_1;
+  }
+  if (min > TLS1_2_VERSION) {
+    opts |= SSL_OP_NO_TLSv1_2;
+  }
+
+  if (max < TLS1_2_VERSION) {
+    opts |= SSL_OP_NO_TLSv1_2;
+  }
+  if (max < TLS1_1_VERSION) {
+    opts |= SSL_OP_NO_TLSv1_1;
+  }
+
+  SSL_CTX_set_options(ssl_ctx, opts);
+
+  return 0;
+#endif // !OPENSSL_1_1_API && !defined(OPENSSL_IS_BORINGSSL)
 }
 
 } // namespace ssl
